
/* ====================================================================
 * Copyright (c) 1996 The Apache Group.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer. 
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. All advertising materials mentioning features or use of this
 *    software must display the following acknowledgment:
 *    "This product includes software developed by the Apache Group
 *    for use in the Apache HTTP server project (http://www.apache.org/)."
 *
 * 4. The names "Apache Server" and "Apache Group" must not be used to
 *    endorse or promote products derived from this software without
 *    prior written permission.
 *
 * 5. Redistributions of any form whatsoever must retain the following
 *    acknowledgment:
 *    "This product includes software developed by the Apache Group
 *    for use in the Apache HTTP server project (http://www.apache.org/)."
 *
 * THIS SOFTWARE IS PROVIDED BY THE APACHE GROUP ``AS IS'' AND ANY
 * EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE APACHE GROUP OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of the Apache Group and was originally based
 * on public domain software written at the National Center for
 * Supercomputing Applications, University of Illinois, Urbana-Champaign.
 * For more information on the Apache Group and the Apache HTTP server
 * project, please see <http://www.apache.org/>.
 *
 */

/*
**  mod_rewrite.c: URI rewriting module 
**  Written by Ralf S. Engelschall <rse@en.muc.de> 
**
**  HISTORY
**      V1.0 950201 *GENESIS* 
**           950205 - works fine with Apache 1.0.2
**           950210 - adapted for Apache 1.1b0a
**      V1.1 950229 - added comments and made ready for distribution
**                  - fixed a bug when run under Solaris 2.5:
**                    a result of OK crashed apache 1.1b0a, DECLINED is ok..
**      V1.2 950229 - released it to ftp://hyperreal.com/httpd/incoming/
**      V1.3 950301 - adapted it for Apache 1.1b0+ak
**           950306 - added [x] flags part to RewriteRule command
**                    and implemented the first flag: 'R' to
**                    force a redirect!
**                  - cleaned up source (now "gcc -Wall" is quiet)
**           950314 - added LASTRULE and NEWROUND flags for rewriting
**                  - added support for QUERY_STRINGS which are
**                    generated by a rewriting rule 
**           950317 - released it on EVA/FAUS
**      V1.4 950328 - added SCRIPT_URL and SCRIPT_URI environment variables
**                    which hold the un-rewritten URL and URI from the
**                    request. Use this in the footer of your pages with SSI.
**           950405 - released it on EVA/FAUS
**      V1.5 950423 - lots of little comment changes to make the module clean
**                  - tested under Apache 1.1b1
**                  - made gcc -Wall happy
**                  - released it to ftp://playground.sdm.de/pub/apache/mod_rewrite/
**
**  INSTALLATION
**
**      o copy files to apache distribution sourcetree:
**
**        $ cp    mod_rewrite.c %DISTROOT%/src/
**        $ cp -r regexp        %DISTROOT%/src/
**
**      o add the following lines to your %DISTROOT%/src/Configuration:
**
**        |    :
**        | EXTRA_LIBS= regexp/libregexp.a
**        |    :
**        | Module rewrite_module mod_rewrite.c
**        |    :
**
**      o precompile the additional libraries (because
**        I don't want to change the Apache makefiles):
**
**        $ (cd regexp; make)
**
**      o now go on and compile Apache as usual...
**
**        $ ./Configure
**        $ make
**
**  CONFIGURATION (inside httpd.conf)
**
**      RewriteLog  <filename>
**      RewriteMap  <mapname>   [<type>:]<filename>
**      RewriteRule <regexp>    <regsub>
**
**      with 
**
**      <filename> ::= ``an absolute Unix pathname to a logfile''
**      <mapname>  ::= [a-zA-Z0-9]+
**      <type>     ::= txt|dbm
**      <regexp>   ::= ``a Unix V8 regular expression for the input URI''
**      <regsub>   ::= ``a output URI with \0 - \9 expanded to the
**                       regexp-bracket-parts and
**                       "${"<mapname>":"<key>["|"<default>"}" expanded
**                       to the value corresponding to <key> in <mapname>
**                       or <default> if <key> was not found in <mapname>''
**                     . "[" <flags> "]"
**      <flags>    ::= "R" |   ``force redirect, i.e. prefix output with
**                               the string http://thishost/''
**                     "L" |   ``last rule, i.e. stop rewriting here''
**                     "N"     ``new round, i.e. stop this round here and begin 
**                               again with first rule''
**
*/

#define ENVVAR_SCRIPT_URL "SCRIPT_URL"
#define ENVVAR_SCRIPT_URI "SCRIPT_URI"

#define DEFAULT_REWRITELOGFILE "/var/log/apache/rewrite.log"

#ifndef SUPPORT_DBM_REWRITEMAP
#define SUPPORT_DBM_REWRITEMAP 0
#endif

#include <string.h>
#include <stdarg.h>

#include "httpd.h"
#include "http_config.h"
#include "http_core.h"
#include "http_log.h"

#include "regexp/regexp.h"
#if SUPPORT_DBM_REWRITEMAP
#include "sdbm/sdbm.h"
#endif

/*
**
**  our private data structures we handle with
**
*/

typedef struct {
    char *name;
    char *file;
} rewritemap_entry;

typedef struct {
    char *pattern;
    regexp *regexp;
    char *output;
    int flags;
} rewriterule_entry;

#define REWRITEFLAG_NONE               0
#define REWRITEFLAG_FORCEREDIRECT      1  /* R */
#define REWRITEFLAG_LASTRULE           2  /* L */
#define REWRITEFLAG_NEWROUND           4  /* N */

typedef struct {
    char *rewritelogfile;       /* the RewriteLog filename */
    int rewritelogfp;           /* the RewriteLog open filepointer */
    array_header *rewritemaps;  /* the RewriteMap entries */
    array_header *rewriterules; /* the RewriteRuke entries */
} rewrite_server_conf;


/*
**
**  forward declarations
**
*/

command_rec rewrite_cmds[];

void *create_rewrite_config(pool *p, server_rec *s);
void *merge_rewrite_config(pool *p, void *basev, void *overridesv);
char *set_rewritelog(cmd_parms *cmd, void *dummy, char *a1);
char *add_rewritemap(cmd_parms *cmd, void *dummy, char *a1, char *a2);
char *add_rewriterule(cmd_parms *cmd, void *dummy, char *a1, char *a2);

int rewrite_apply(request_rec *r);
int apply_rewrite_list(request_rec *r, array_header *rewriterules, int doesc);
int apply_rewrite_rule(request_rec *r, rewriterule_entry *p);

void reduce_uri(request_rec *r);
void splitout_queryargs(request_rec *r);
int is_this_our_host(request_rec *r, char *testhost);
char **make_hostname_list(request_rec *r, char *hostname);

void expand_map_lookups(request_rec *r, char *uri);
char *lookup_map(request_rec *r, char *name, char *key);
char *lookup_map_txtfile(request_rec *r, char *file, char *key);
#if SUPPORT_DBM_REWRITEMAP
char *lookup_map_dbmfile(request_rec *r, char *file, char *key);
#endif

void init_rewritelog(server_rec *s, pool *p);
void open_rewritelog(server_rec *s, pool *p);
void rewritelog_child(void *cmd);
void rewritelog(request_rec *r, char *text, ...);
char *current_logtime(request_rec *r);


/*
**
**  our interface to the Apache server kernel
**
*/

module rewrite_module = {
   STANDARD_MODULE_STUFF,
   init_rewritelog,             /* initializer */
   NULL,                        /* dir config creater */
   NULL,                        /* dir merger --- default is to override */
   create_rewrite_config,       /* server config */
   merge_rewrite_config,        /* merge server configs */
   rewrite_cmds,                /* command table */
   NULL,                        /* handlers */
   rewrite_apply,               /* filename translation */
   NULL,                        /* check_user_id */
   NULL,                        /* check auth */
   NULL,                        /* check access */
   NULL,                        /* type_checker */
   NULL,                        /* fixups */
   NULL                         /* logger */
};

command_rec rewrite_cmds[] = {
    { "RewriteLog", set_rewritelog, NULL, RSRC_CONF, TAKE1, 
      "the filename to the rewriting logfile"},
    { "RewriteMap", add_rewritemap, NULL, RSRC_CONF, TAKE2, 
      "a mapname and a filename"},
    { "RewriteRule", add_rewriterule, NULL, RSRC_CONF, TAKE2, 
      "a input-URL and a output-URL"},
    { NULL }
};


/*
**
**  configuration file stuff
**
*/

void *create_rewrite_config(pool *p, server_rec *s)
{
    rewrite_server_conf *a;
    a = (rewrite_server_conf *)pcalloc(p, sizeof(rewrite_server_conf));
    a->rewritelogfile = DEFAULT_REWRITELOGFILE;
    a->rewritelogfp = -1;
    a->rewritemaps  = make_array(p, 20, sizeof(rewritemap_entry));
    a->rewriterules = make_array(p, 20, sizeof(rewriterule_entry));
    return (void *)a;
}

void *merge_rewrite_config(pool *p, void *basev, void *overridesv)
{
    rewrite_server_conf *a, *base, *overrides;
    a         = (rewrite_server_conf *)pcalloc(p, sizeof(rewrite_server_conf));
    base      = (rewrite_server_conf *)basev;
    overrides = (rewrite_server_conf *)overridesv;
    a->rewritelogfile = overrides->rewritelogfile;
    a->rewritelogfp   = overrides->rewritelogfp;
    a->rewritemaps  = append_arrays(p, overrides->rewritemaps,  base->rewritemaps);
    a->rewriterules = append_arrays(p, overrides->rewriterules, base->rewriterules);
    return (void *)a;
}

char *set_rewritelog(cmd_parms *cmd, void *dummy, char *a1)
{
    rewrite_server_conf *conf;

    conf = (rewrite_server_conf *)get_module_config(cmd->server->module_config, &rewrite_module);
    conf->rewritelogfile = a1;
    return NULL;
}

char *add_rewritemap(cmd_parms *cmd, void *dummy, char *a1, char *a2)
{
    server_rec *s;
    rewrite_server_conf *conf;
    rewritemap_entry *new;

    s = cmd->server;
    conf = (rewrite_server_conf *)get_module_config(s->module_config, &rewrite_module);
    new = push_array(conf->rewritemaps);
    new->name = a1;
    new->file = a2;
    return NULL;
}

char *add_rewriterule(cmd_parms *cmd, void *dummy, char *a1, char *a2)
{
    server_rec *s;
    rewrite_server_conf *conf;
    rewriterule_entry *new;
    regexp *regexp;
    int flags;
    char *cp;

    s = cmd->server;
    conf = (rewrite_server_conf *)get_module_config(s->module_config, &rewrite_module);

    if ((regexp = regcomp(a1)) == NULL) {
        fprintf(stderr, "RewriteRule: cannot compile regular expression '%s'\n", a1);
        exit(1);
    }

    /* split away the ...[<flags>] part */
    flags = REWRITEFLAG_NONE;
    if (a2[strlen(a2)-1] == ']') {
        if ((cp = strrchr(a2, '[')) != NULL) {
            *(cp++) = '\0';
            for ( ; *cp != ']'; cp++) {
                if (*cp == ',' || *cp == ' ' || *cp == '\t' || *cp == ':')
                    continue;
                switch (*cp) {
                    case 'R': flags |= REWRITEFLAG_FORCEREDIRECT; break;
                    case 'L': flags |= REWRITEFLAG_LASTRULE;      break;
                    case 'N': flags |= REWRITEFLAG_NEWROUND;      break;
                    default: 
                        fprintf(stderr, "RewriteRule: unkown flag '%c'\n", *cp);
                        exit(1);
                }
            }
        }
    }
    new = push_array(conf->rewriterules);
    new->pattern = a1;
    new->regexp  = regexp;
    new->output  = a2;
    new->flags   = flags;
    return NULL;
}


/*
**
**  RewriteRule support
**
*/

int rewrite_apply(request_rec *r)
{
    void *sconf;
    rewrite_server_conf *conf;
    char *var;
    char *thisserver, *thisport, *thisurl;
    char buf[512];

    sconf = r->server->module_config;
    conf  = (rewrite_server_conf *)get_module_config(sconf, &rewrite_module);

    /* add the SCRIPT_URL variable to the env. this is a bit complicated
       due to the fact that apache uses subrequests and internal redirects */
    if (r->main == NULL) {
         var = pstrcat(r->pool, "REDIRECT_", ENVVAR_SCRIPT_URL, NULL);
         var = table_get(r->subprocess_env, var);
         if (var == NULL) 
             table_set(r->subprocess_env, ENVVAR_SCRIPT_URL, pstrdup(r->pool, r->uri));
         else 
             table_set(r->subprocess_env, ENVVAR_SCRIPT_URL, pstrdup(r->pool, var));
    } 
    else {
         var = table_get(r->main->subprocess_env, ENVVAR_SCRIPT_URL);
         table_set(r->subprocess_env, ENVVAR_SCRIPT_URL, pstrdup(r->pool, var));
    }

    /* add the canonical URI of this URL */
    thisserver = r->server->server_hostname;
    if (r->server->port == 80) 
        thisport = "";
    else {
        sprintf(buf, ":%d", r->server->port);
        thisport = pstrdup(r->pool, buf);
    }
    thisurl = table_get(r->subprocess_env, ENVVAR_SCRIPT_URL);
    var = pstrcat(r->pool, "http://", thisserver, thisport, thisurl, NULL);
    table_set(r->subprocess_env, ENVVAR_SCRIPT_URI, pstrdup(r->pool, var));

    /* now apply the rules ... */
    if (apply_rewrite_list(r, conf->rewriterules, 0)) {

        if (strncmp(r->uri, "http://", 7) == 0) {
            /* it was finally rewritten to a remote path */
            rewritelog(r, "redirect to %s", r->uri);
            table_set(r->headers_out, "Location", r->uri);
            return REDIRECT;
        }
        else {
            /* it was finally rewritten to a local path */
            rewritelog(r, "go-ahead with %s", r->uri);
            r->filename = pstrdup(r->pool, r->uri);
            /* Hhhmmmmmmmmm....
               I don't understand it in the deep:
               Under a FreeBSD 2.1.0 box, a result of OK works fine,
               under a Solaris 2.5 box, OK gives a Coredump whereas DECLINED works!
               WHAT THE HELL IS CORRECT??????? I think DECLINED.... but who knows? */
            return DECLINED;
        }
    }
    else {
        rewritelog(r, "pass through %s", r->uri);
        return DECLINED;
    }
}

int apply_rewrite_list(request_rec *r, array_header *rewriterules, int doesc)
{
    rewriterule_entry *entries;
    rewriterule_entry *p;
    int i;
    int changed;
    
    entries = (rewriterule_entry *)rewriterules->elts;
    changed = 0;
    loop:
    for (i = 0; i < rewriterules->nelts; i++) {
        p = &entries[i];
        if (apply_rewrite_rule(r, p)) {
            changed = 1;
            if (p->flags & REWRITEFLAG_LASTRULE) 
                break;
            if (p->flags & REWRITEFLAG_NEWROUND) 
                goto loop;
        }
    }
    return changed;
}

int apply_rewrite_rule(request_rec *r, rewriterule_entry *p)
{
    int changed;
    char *uri;
    char *output;
    int flags;
    char newuri[MAX_STRING_LEN];
    regexp *regexp;

    uri     = r->uri;
    regexp  = p->regexp;
    output  = p->output;
    flags   = p->flags;

    changed = 0;
    if (regexec(regexp, uri)) {                              /* try to match the pattern */

        /* apply the rewrite output */
        regsub(regexp, output, newuri);                      /* subsituate in output */
        expand_map_lookups(r, newuri);                       /* expand ${...} */
        rewritelog(r, "rewrite %s -> %s", r->uri, newuri);
        r->uri = pstrdup(r->pool, newuri);

        /* reduce http://<ourhost> */
        reduce_uri(r);

        /* split out on-the-fly generated QUERY_STRING '....?xxxxx&xxxx...' */
        splitout_queryargs(r);

        /* if we are forced to do a explicit redirect
           finally prefix the new URI with http://<ourname> explicitly */
        if (flags & REWRITEFLAG_FORCEREDIRECT) {
            if (strncmp(r->uri, "http://", 7) != 0) {
                sprintf(newuri, "http://%s%s", r->server->server_hostname, r->uri);
                rewritelog(r, "prepare forced redirect %s -> %s", r->uri, newuri);
                r->uri = pstrdup(r->pool, newuri);
            }
        }
        changed = 1;
    }
    return changed;
}



/*
**
**  DNS hostname check
**
*/

int is_this_our_host(request_rec *r, char *testhost)
{
    char **cppHNLour;
    char **cppHNLtest;
    int i, j;

    if ((cppHNLour = make_hostname_list(r, r->server->server_hostname)) == NULL)
        return 0;
    if ((cppHNLtest = make_hostname_list(r, testhost)) == NULL)
        return 0;

    for (i = 0; cppHNLtest[i] != NULL; i++) {
        for (j = 0; cppHNLour[j] != NULL; j++) {
            if (strcmp(cppHNLtest[i], cppHNLour[j]) == 0) {
                return 1;
            }
        }
    }
    return 0;
}

char **make_hostname_list(request_rec *r, char *hostname)
{
    char **cppHNL;
    struct hostent *hep;
    int i;

    if ((hep = gethostbyname(hostname)) == NULL)
        return NULL;
    for (i = 0; hep->h_aliases[i]; i++)
        ;
    cppHNL = (char **)palloc(r->pool, sizeof(char *)*(i+2));
    cppHNL[0] = pstrdup(r->pool, hep->h_name);
    for (i = 0; hep->h_aliases[i]; i++)
        cppHNL[1+i] = pstrdup(r->pool, hep->h_aliases[i]);
    cppHNL[1+i] = NULL;
    return cppHNL;
}

void reduce_uri(request_rec *r)
{
    char *a;
    char *e;
    int n;
    char hostname[MAX_STRING_LEN];
    char olduri[MAX_STRING_LEN];

    if (strncmp(r->uri, "http://", 7) == 0) {
        /* there was a rewrite to a remote path */

        strcpy(olduri, r->uri); /* save for logging */

        /* cut the hostname out of the URI */
        a = r->uri+7;
        e = strchr(a, '/');
        if (e == NULL)
            e = a+strlen(a);
        n = e-a;
        memcpy(hostname, a, n);
        hostname[n] = '\0';

        /* check whether we could reduce it to a local path... */
        if (is_this_our_host(r, hostname)) {
            /* this is our host, so remove http://ourhostname/ */
            if (*e == '\0')
                r->uri = pstrdup(r->pool, "/");
            else
                r->uri = pstrdup(r->pool, e);
            rewritelog(r, "reduce %s -> %s", olduri, r->uri);
        }
    }
    return;            
}

void splitout_queryargs(request_rec *r)
{
    char *q;
    char olduri[MAX_STRING_LEN];

    q = strchr(r->uri, '?');
    if (q != NULL) {
        strcpy(olduri, r->uri);
        *q++ = '\0';
        r->args = pstrcat(r->pool, q, "&", r->args, NULL);
        rewritelog(r, "split %s -> %s, args=%s", olduri, r->uri, r->args);
    }
    return;            
}


/*
**
**  mapfile expansion support
**  i.e. expansion of MAP lookup directives
**  ${<mapname>:<key>} in RewriteRule rhs
**
*/

void expand_map_lookups(request_rec *r, char *uri)
{
    char newuri[MAX_STRING_LEN];
    char *cpI;
    char *cpIE;
    char *cpO;
    char *cpT;
    char *cpT2;
    char mapname[MAX_STRING_LEN];
    char mapkey[MAX_STRING_LEN];
    char defaultvalue[MAX_STRING_LEN];
    int n;

    cpI = uri;
    cpIE = cpI+strlen(cpI);
    cpO = newuri;
    while (cpI < cpIE) {
        if (cpI+6 < cpIE && strncmp(cpI, "${", 2) == 0) {
            /* missing delimiter -> take it as plain text */
            if (   strchr(cpI+2, ':') == NULL
                || strchr(cpI+2, '}') == NULL) {
                memcpy(cpO, cpI, 2);
                cpO += 2;
                cpI += 2;
                continue;
            }
            cpI += 2;

            cpT = strchr(cpI, ':');
            n = cpT-cpI;
            memcpy(mapname, cpI, n);
            mapname[n] = '\0';
            cpI += n+1;

            cpT2 = strchr(cpI, '|');
            cpT = strchr(cpI, '}');
            if (cpT2 != NULL && cpT2 < cpT) {
                n = cpT2-cpI;
                memcpy(mapkey, cpI, n);
                mapkey[n] = '\0';
                cpI += n+1;

                n = cpT-cpI;
                memcpy(defaultvalue, cpI, n);
                defaultvalue[n] = '\0';
                cpI += n+1;
            }
            else {
                n = cpT-cpI;
                memcpy(mapkey, cpI, n);
                mapkey[n] = '\0';
                cpI += n+1;

                defaultvalue[0] = '\0';
            }

            cpT = lookup_map(r, mapname, mapkey);
            if (cpT != NULL) {
                n = strlen(cpT);
                memcpy(cpO, cpT, n);
                cpO += n;
            }
            else {
                n = strlen(defaultvalue);
                memcpy(cpO, defaultvalue, n);
                cpO += n;
            }
        }
        else {
            cpT = strstr(cpI, "${");
            if (cpT == NULL)
                cpT = cpI+strlen(cpI);
            n = cpT-cpI;
            memcpy(cpO, cpI, n);
            cpO += n;
            cpI += n;
        }
    }
    *cpO = '\0';
    strcpy(uri, newuri);
    return;
}

char *lookup_map(request_rec *r, char *name, char *key)
{
    void *sconf;
    rewrite_server_conf *conf;
    array_header *rewritemaps;
    rewritemap_entry *entries;
    rewritemap_entry *s;
    char *value;
    int i;

    /* get map configuration */
    sconf = r->server->module_config;
    conf  = (rewrite_server_conf *)get_module_config(sconf, &rewrite_module);
    rewritemaps = conf->rewritemaps;

    entries = (rewritemap_entry *)rewritemaps->elts;
    for (i = 0; i < rewritemaps->nelts; i++) {
        s = &entries[i];
        if (strcmp(s->name, name) == 0) {
            if (strlen(s->file) > 4 && strncmp(s->file, "txt:", 4) == 0) {
                if ((value = lookup_map_txtfile(r, s->file+4, key)) != NULL)
                    return value;
            }
            else if (strlen(s->file) > 4 && strncmp(s->file, "dbm:", 4) == 0) {
#if SUPPORT_DBM_REWRITEMAP
                if ((value = lookup_map_dbmfile(r, s->file+4, key)) != NULL)
                    return value;
#else
                return NULL;
#endif
            }
            else {
                if ((value = lookup_map_txtfile(r, s->file, key)) != NULL)
                    return value;
            }
        }
    }
    return NULL;
}

#define PATTERN "^([^ ]+) +([^ ]+).*$"
#define OUTPUT  "\\1,\\2"
regexp *lookup_map_txtfile_regexp = NULL;

char *lookup_map_txtfile(request_rec *r, char *file, char *key)
{
    FILE *fp = NULL;
    char line[1024];
    char output[1024];
    char result[1024];
    char *value = NULL;
    char *cpT;
    char *curkey;
    char *curval;

    if ((fp = pfopen(r->pool, file, "r")) == NULL)
        return NULL;

    if (lookup_map_txtfile_regexp == NULL)
        lookup_map_txtfile_regexp = regcomp(PATTERN);

    strcpy(output,  OUTPUT);
    while (fgets(line, sizeof(line), fp) != NULL) {
        if (line[strlen(line)-1] == '\n')
            line[strlen(line)-1] = '\0';
        if (regexec(lookup_map_txtfile_regexp, line)) {
            regsub(lookup_map_txtfile_regexp, output, result);
            cpT = strchr(result, ',');
            *cpT = '\0';
            curkey = result;
            curval = cpT+1;

            if (strcmp(curkey, key) == 0) {
                value = pstrdup(r->pool, curval);
                break;
            }
        }
    }
    pfclose(r->pool, fp);
    return value;
}

#if SUPPORT_DBM_REWRITEMAP
char *lookup_map_dbmfile(request_rec *r, char *file, char *key)
{
    DBM *dbmfp = NULL;
    datum dbmkey;
    datum dbmval;
    char *value = NULL;
    char buf[MAX_STRING_LEN];

    dbmkey.dptr  = key;
    dbmkey.dsize = strlen(key)+1;
    if ((dbmfp = dbm_open(file, O_RDONLY, 0666)) != NULL) {
        dbmval = dbm_fetch(dbmfp, dbmkey);
        if (dbmval.dptr != NULL) {
            memcpy(buf, dbmval.dptr, dbmval.dsize);
            buf[dbmval.dsize] = '\0';
            value = pstrdup(r->pool, buf);
        }
        dbm_close(dbmfp);
    }
    return value;
}
#endif


/*
**
**  RewriteLog support
**
*/

void init_rewritelog(server_rec *s, pool *p)
{
    for (; s; s = s->next)
        open_rewritelog(s, p);
}

static int    rewritelog_flags = ( O_WRONLY|O_APPEND|O_CREAT );
static mode_t rewritelog_mode  = ( S_IRUSR|S_IWUSR|S_IRGRP|S_IROTH );

void open_rewritelog(server_rec *s, pool *p)
{
    rewrite_server_conf *conf;
    char *fname;
    FILE *fp;
  
    conf = get_module_config(s->module_config, &rewrite_module);
    
    if (conf->rewritelogfile == NULL)
        return; 
    if (*(conf->rewritelogfile) == '\0')
        return;
    if (conf->rewritelogfp > 0)
        return; /* virtual log shared w/main server */

    fname = server_root_relative(p, conf->rewritelogfile);
    
    if (*conf->rewritelogfile == '|') {
        spawn_child(p, rewritelog_child, (void *)(conf->rewritelogfile+1),
                    kill_after_timeout, &fp, NULL);
        if (fp == NULL) {
            fprintf (stderr, "Couldn't fork child for RewriteLog process\n");
            exit (1);
        }
        conf->rewritelogfp = fileno(fp);
    }
    else if (*conf->rewritelogfile != '\0') {
        if ((conf->rewritelogfp = popenf(p, fname, rewritelog_flags, rewritelog_mode)) < 0) {
            fprintf(stderr, "httpd: could not open rewrite log file %s.\n", fname);
            perror("open");
            exit(1);
        }
    }
}

/* Child process code for 'RewriteLog "|..."' */
void rewritelog_child(void *cmd)
{
    cleanup_for_exec();
    signal(SIGHUP, SIG_IGN);
    execl(SHELL_PATH, SHELL_PATH, "-c", (char *)cmd, NULL);
    exit(1);
}

void rewritelog(request_rec *r, char *text, ...)
{
    rewrite_server_conf *conf;
    conn_rec *connect;
    char *str1;
    char str2[HUGE_STRING_LEN];
    char str3[HUGE_STRING_LEN];
    va_list ap;
    
    va_start(ap, text);
    conf = get_module_config(r->server->module_config, &rewrite_module);
    connect = r->connection;

    if (conf->rewritelogfp <0)
        return;
    if (conf->rewritelogfile == NULL)
        return;
    if (*(conf->rewritelogfile) == '\0')
        return;

    str1 = pstrcat(r->pool, get_remote_host(connect, r->server->module_config, REMOTE_NAME), " ",
                            (connect->remote_logname != NULL ? connect->remote_logname : "-"), " ",
                            (connect->user != NULL ? connect->user : "-"),
                            NULL);
    vsprintf(str2, text, ap);
    sprintf(str3, "%s %s [%x] %s\n", str1, current_logtime(r), (unsigned int)r, str2);

    write(conf->rewritelogfp, str3, strlen(str3));

    va_end(ap);
    return;
}

char *current_logtime(request_rec *r)
{
    long timz;
    struct tm *t;
    char tstr[MAX_STRING_LEN], sign;
    
    t = get_gmtoff(&timz);
    sign = (timz < 0 ? '-' : '+');
    if(timz < 0) 
        timz = -timz;

    strftime(tstr, MAX_STRING_LEN,"[%d/%b/%Y:%H:%M:%S ",t);

    sprintf (tstr + strlen(tstr), "%c%02ld%02ld]", sign, timz/3600, timz%3600);

    return pstrdup(r->pool, tstr);
}


/**EOF**/
