
/* ====================================================================
 * Copyright (c) 1996 The Apache Group.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer. 
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. All advertising materials mentioning features or use of this
 *    software must display the following acknowledgment:
 *    "This product includes software developed by the Apache Group
 *    for use in the Apache HTTP server project (http://www.apache.org/)."
 *
 * 4. The names "Apache Server" and "Apache Group" must not be used to
 *    endorse or promote products derived from this software without
 *    prior written permission.
 *
 * 5. Redistributions of any form whatsoever must retain the following
 *    acknowledgment:
 *    "This product includes software developed by the Apache Group
 *    for use in the Apache HTTP server project (http://www.apache.org/)."
 *
 * THIS SOFTWARE IS PROVIDED BY THE APACHE GROUP ``AS IS'' AND ANY
 * EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE APACHE GROUP OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of the Apache Group and was originally based
 * on public domain software written at the National Center for
 * Supercomputing Applications, University of Illinois, Urbana-Champaign.
 * For more information on the Apache Group and the Apache HTTP server
 * project, please see <http://www.apache.org/>.
 *
 */


/*
 * Dynamic loading using the dlopen() and dlsym() functions.
 * Yves Arrouye <arrouye@debian.org>, based on mod_dld.c.
 *
 * CAVEATS: this module currently uses a hack to know wether to load
 * files or not. Since I don not have precise informations about how
 * to manage memory in a module, I cannot do the following, but I
 * would like infos to do it or to see someone do it:
 *
 *   - register which modules/files have been handled by LoadModule or
 *     LoadFile, so that when Apache reloads its configuration newly
 *     added modules can be loaded (currently one must stop and
 *     restart the server).
 *
 *   - add a count so that we can also determine that a module/file
 *     load directive has been removed, in which case the file must be
 *     removed too.
 *
 */

#include "httpd.h"
#include "http_config.h"
#include "http_conf_globals.h"	/* server_argv0.  Sigh... */
#include <dlfcn.h>

/*
 * The hard part of implementing LoadModule is deciding what to do about
 * rereading the config files.  This proof-of-concept implementation takes the 
 * cheap way out:  we only actually load the modules the first time through.
 */

static int been_there_done_that = 0; /* Loaded the modules yet? */

char *link_file (pool *p, char *filename, void** handle)
{
    filename = server_root_relative (p, filename);
    if ((*handle = dlopen(filename, RTLD_LAZY)) == 0) {
	fprintf(stderr, "Error loading %s: %s.\n", filename,
	    dlerror());
	return pstrcat (p, "Cannot load ", filename, " into server", NULL);
    }
    return NULL;
}

char *load_module (cmd_parms *cmd, void *dummy, char *modname, char *filename)
{
    void* handle;
    char *errname;
    module *modp;

    extern char* module_names[];
    extern int module_hmany;

    if (been_there_done_that) return NULL;
    
    if ((errname = link_file (cmd->pool, filename, &handle))) return errname;
    if (!(modp = (module *)dlsym (handle, modname))) {
	return pstrcat (cmd->pool, "Can't find module ", modname,
			           " in file ", filename, NULL);
    }

    add_module (modp);
    module_names[module_hmany++] = strdup(modname);

    /* Alethea Patch (rws,djw2) - need to run configuration functions
       in new modules */

    if (modp->create_server_config)
      ((void**)cmd->server->module_config)[modp->module_index]=
	(*modp->create_server_config)(cmd->pool, cmd->server);

    if (modp->create_dir_config)
      ((void**)cmd->server->lookup_defaults)[modp->module_index]=
	(*modp->create_dir_config)(cmd->pool, NULL);


    return NULL;
}

char *load_file (cmd_parms *cmd, void *dummy, char *filename)
{
    void* handle;
    char *errname;
    
    if (been_there_done_that) return NULL;
    
    errname = link_file (cmd->pool, filename, &handle);
    if (handle) dlclose(handle);

    return errname;
}

void say_loaded_modules (server_rec *dummy, pool *p)
{
    been_there_done_that = 1;
    return;
}

command_rec dld_cmds[] = {
{ "LoadModule", load_module, NULL, RSRC_CONF, TAKE2,
  "a module name, and the name of a file to load it from"},
{ "LoadFile", load_file, NULL, RSRC_CONF, ITERATE,
  "files or libraries to link into the server at runtime"},
{ NULL }
};

module dl_module = {
   STANDARD_MODULE_STUFF,
   say_loaded_modules,		/* initializer */
   NULL,			/* create per-dir config */
   NULL,			/* merge per-dir config */
   NULL,			/* server config */
   NULL,			/* merge server config */
   dld_cmds,			/* command table */
   NULL,			/* handlers */
   NULL,			/* filename translation */
   NULL,			/* check_user_id */
   NULL,			/* check auth */
   NULL,			/* check access */
   NULL,			/* type_checker */
   NULL				/* logger */
};

